#include "dfsch-ext/cairo.h"
#include <dfsch/dfsch.h>

typedef struct context_t {
  dfsch_type_t* type;
  cairo_t* cr;
} context_t;

dfsch_type_t dfsch_ext_cairo_context_type = {
  DFSCH_STANDARD_TYPE,
  NULL,
  sizeof(context_t),
  "cairo:context",
  NULL,
  NULL,
  NULL,
  NULL
};

static void context_finalizer(context_t* cairo, void* cd){
  if (cairo->cr){
    cairo_destroy(cairo->cr);
    cairo->cr = NULL;
  }
}

dfsch_object_t* dfsch_ext_cairo_make_context(cairo_t* cr){
  dfsch_object_t* cairo = dfsch_make_object(DFSCH_EXT_CAIRO_CONTEXT_TYPE);

  GC_REGISTER_FINALIZER(cairo, (GC_finalization_proc)context_finalizer,
                        NULL, NULL, NULL);
  ((context_t*)cairo)->cr = cr;

  return cairo;
}

cairo_t* dfsch_ext_cairo_context(dfsch_object_t* cairo){
  if (DFSCH_TYPE_OF(cairo) != DFSCH_EXT_CAIRO_CONTEXT_TYPE){
    dfsch_error("cairo:not-a-context", cairo);
  }

  if (!((context_t*)cairo)->cr){
    dfsch_error("cairo:context-already-destroyed", cairo);
  }

  return ((context_t*)cairo)->cr;
}

void dfsch_ext_cairo_destroy_context(dfsch_object_t* cairo){
  if (DFSCH_TYPE_OF(cairo) != DFSCH_EXT_CAIRO_CONTEXT_TYPE){
    dfsch_error("cairo:not-a-context", cairo);
  }

  if (!((context_t*)cairo)->cr){
    dfsch_error("cairo:context-already-destroyed", cairo);
  }

  cairo_destroy(((context_t*)cairo)->cr);
  ((context_t*)cairo)->cr = NULL;
}

typedef struct surface_t {
  dfsch_type_t* type;
  cairo_surface_t* surface;
} surface_t;

dfsch_type_t dfsch_ext_cairo_surface_type = {
  DFSCH_ABSTRACT_TYPE,
  NULL,
  sizeof(surface_t),
  "cairo:surface",
  NULL,
  NULL,
  NULL,
  NULL
};

dfsch_type_t dfsch_ext_cairo_image_surface_type = {
  DFSCH_STANDARD_TYPE,
  DFSCH_EXT_CAIRO_SURFACE_TYPE,
  sizeof(surface_t),
  "cairo:image-surface",
  NULL,
  NULL,
  NULL,
  NULL
};
#ifdef CAIRO_HAS_PDF_SURFACE
dfsch_type_t dfsch_ext_cairo_pdf_surface_type = {
  DFSCH_STANDARD_TYPE,
  DFSCH_EXT_CAIRO_SURFACE_TYPE,
  sizeof(surface_t),
  "cairo:pdf-surface",
  NULL,
  NULL,
  NULL,
  NULL
};
#endif
#ifdef CAIRO_HAS_PS_SURFACE
dfsch_type_t dfsch_ext_cairo_ps_surface_type = {
  DFSCH_STANDARD_TYPE,
  DFSCH_EXT_CAIRO_SURFACE_TYPE,
  sizeof(surface_t),
  "cairo:ps-surface",
  NULL,
  NULL,
  NULL,
  NULL
};
#endif
#ifdef CAIRO_HAS_SVG_SURFACE
dfsch_type_t dfsch_ext_cairo_svg_surface_type = {
  DFSCH_STANDARD_TYPE,
  DFSCH_EXT_CAIRO_SURFACE_TYPE,
  sizeof(surface_t),
  "cairo:svg-surface",
  NULL,
  NULL,
  NULL,
  NULL
};
#endif
#ifdef CAIRO_HAS_XLIB_SURFACE
dfsch_type_t dfsch_ext_cairo_xlib_surface_type = {
  DFSCH_STANDARD_TYPE,
  DFSCH_EXT_CAIRO_SURFACE_TYPE,
  sizeof(surface_t),
  "cairo:xlib-surface",
  NULL,
  NULL,
  NULL,
  NULL
};
#endif


static void surface_finalizer(surface_t* surface, void* cd){
  if (surface->surface){
    cairo_surface_destroy(surface->surface);
    surface->surface = NULL;
  }
}

dfsch_object_t* dfsch_ext_cairo_make_surface(cairo_surface_t* surface,
                                             dfsch_type_t* type){
  dfsch_object_t* s = dfsch_make_object(type);

  GC_REGISTER_FINALIZER(s, (GC_finalization_proc)surface_finalizer,
                        NULL, NULL, NULL);
  ((surface_t*)s)->surface = surface;

  return s;
}

cairo_surface_t* dfsch_ext_cairo_surface(dfsch_object_t* surface){
  if (!DFSCH_INSTANCE_P(surface, DFSCH_EXT_CAIRO_SURFACE_TYPE)){
    dfsch_error("cairo:not-a-surface", surface);
  }

  if (!((surface_t*)surface)->surface){
    dfsch_error("cairo:surface-already-destroyed", surface);
  }

  return ((surface_t*)surface)->surface;
}

void dfsch_ext_cairo_destroy_surface(dfsch_object_t* surface){
  if (!DFSCH_INSTANCE_P(surface, DFSCH_EXT_CAIRO_SURFACE_TYPE)){
    dfsch_error("cairo:not-a-surface", surface);
  }

  if (!((surface_t*)surface)->surface){
    dfsch_error("cairo:surface-already-destroyed", surface);
  }

  cairo_surface_destroy(((surface_t*)surface)->surface);
  ((surface_t*)surface)->surface = NULL;
}
