#include "dfsch-ext/cairo.h"
#include <dfsch/dfsch.h>
#include <dfsch/load.h>
#include <dfsch/number.h>

/*
 * Context handling
 */

DFSCH_DEFINE_PRIMITIVE(make_context, 0){
  cairo_surface_t* surface;

  CAIRO_SURFACE_ARG(args, surface);
  DFSCH_ARG_END(args);

  return dfsch_ext_cairo_make_context(cairo_create(surface));
}

DFSCH_DEFINE_PRIMITIVE(destroy_context, 0){
  dfsch_object_t* context;

  DFSCH_OBJECT_ARG(args, context);
  DFSCH_ARG_END(args);

  dfsch_ext_cairo_destroy_context(context);
  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(save_context, 0){
  cairo_t* context;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_ARG_END(args);

  cairo_save(context);
  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(restore_context, 0){
  cairo_t* context;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_ARG_END(args);

  cairo_restore(context);
  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(push_group, 0){
  cairo_t* context;
  cairo_content_t content = CAIRO_CONTENT_COLOR_ALPHA;

  CAIRO_CONTEXT_ARG(args, context);
  CAIRO_CONTENT_ARG_OPT(args, content);
  DFSCH_ARG_END(args);

  cairo_push_group_with_content(context, content);
  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(pop_group, 0){
  cairo_t* context;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_ARG_END(args);

  dfsch_error("cairo:pattern-not-supported", NULL);
  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(pop_group_to_source, 0){
  cairo_t* context;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_ARG_END(args);

  cairo_pop_group_to_source(context);
  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(set_source_color, 0){
  cairo_t* context;
  double r, g, b, a;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_DOUBLE_ARG(args, r);
  DFSCH_DOUBLE_ARG(args, g);
  DFSCH_DOUBLE_ARG(args, b);
  DFSCH_DOUBLE_ARG_OPT(args, a, 1.0);
  DFSCH_ARG_END(args);

  cairo_set_source_rgba(context, r, g, b, a);
  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(set_source, 0){
  dfsch_error("cairo:not-implemented", NULL);
}
DFSCH_DEFINE_PRIMITIVE(set_source_surface, 0){
  cairo_t* context;
  cairo_surface_t* surface;
  double x, y;

  CAIRO_CONTEXT_ARG(args, context);
  CAIRO_SURFACE_ARG(args, surface);
  DFSCH_DOUBLE_ARG(args, x);
  DFSCH_DOUBLE_ARG(args, y);
  DFSCH_ARG_END(args);

  cairo_set_source_surface(context, surface, x, y);
  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(set_antialias, 0){
  cairo_t* context;
  cairo_antialias_t antialias;

  CAIRO_CONTEXT_ARG(args, context);
  CAIRO_ANTIALIAS_ARG(args, antialias);
  DFSCH_ARG_END(args);

  cairo_set_antialias(context, antialias);
  return NULL;
}

// . . .

DFSCH_DEFINE_PRIMITIVE(set_line_width, 0){
  cairo_t* context;
  double width;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_DOUBLE_ARG(args, width);
  DFSCH_ARG_END(args);

  cairo_set_line_width(context, width);
  return NULL;
}

// . . .

DFSCH_DEFINE_PRIMITIVE(paint, 0){
  cairo_t* context;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_ARG_END(args);

  cairo_paint(context);
  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(fill, 0){
  cairo_t* context;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_ARG_END(args);

  cairo_fill(context);
  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(fill_preserve, 0){
  cairo_t* context;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_ARG_END(args);

  cairo_fill_preserve(context);
  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(stroke, 0){
  cairo_t* context;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_ARG_END(args);

  cairo_stroke(context);
  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(stroke_preserve, 0){
  cairo_t* context;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_ARG_END(args);

  cairo_stroke_preserve(context);
  return NULL;
}



/*
 * Paths
 */

DFSCH_DEFINE_PRIMITIVE(current_point, 0){
  cairo_t* context;
  double x;
  double y;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_ARG_END(args);

  cairo_get_current_point(context, &x, &y);
  
  return dfsch_vector(2,
                      dfsch_make_number_from_double(x),
                      dfsch_make_number_from_double(y));
}
DFSCH_DEFINE_PRIMITIVE(new_path, 0){
  cairo_t* context;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_ARG_END(args);

  cairo_new_path(context);

  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(new_sub_path, 0){
  cairo_t* context;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_ARG_END(args);

  cairo_new_sub_path(context);

  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(close_path, 0){
  cairo_t* context;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_ARG_END(args);

  cairo_close_path(context);

  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(arc, 0){
  cairo_t* context;
  double xc, yc, radius, angle1, angle2;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_DOUBLE_ARG(args, xc);
  DFSCH_DOUBLE_ARG(args, yc);
  DFSCH_DOUBLE_ARG(args, radius);
  DFSCH_DOUBLE_ARG(args, angle1);
  DFSCH_DOUBLE_ARG(args, angle2);
  DFSCH_ARG_END(args);

  cairo_arc(context, xc, yc, radius, angle1, angle2);

  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(arc_negative, 0){
  cairo_t* context;
  double xc, yc, radius, angle1, angle2;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_DOUBLE_ARG(args, xc);
  DFSCH_DOUBLE_ARG(args, yc);
  DFSCH_DOUBLE_ARG(args, radius);
  DFSCH_DOUBLE_ARG(args, angle1);
  DFSCH_DOUBLE_ARG(args, angle2);
  DFSCH_ARG_END(args);

  cairo_arc_negative(context, xc, yc, radius, angle1, angle2);

  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(curve_to, 0){
  cairo_t* context;
  double x1, y1, x2, y2, x3, y3;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_DOUBLE_ARG(args, x1);
  DFSCH_DOUBLE_ARG(args, y1);
  DFSCH_DOUBLE_ARG(args, x2);
  DFSCH_DOUBLE_ARG(args, y2);
  DFSCH_DOUBLE_ARG(args, x3);
  DFSCH_DOUBLE_ARG(args, y3);
  DFSCH_ARG_END(args);

  cairo_curve_to(context, x1, y1, x2, y2, x3, y3);

  return NULL;
}

DFSCH_DEFINE_PRIMITIVE(line_to, 0){
  cairo_t* context;
  double x, y;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_DOUBLE_ARG(args, x);
  DFSCH_DOUBLE_ARG(args, y);
  DFSCH_ARG_END(args);

  cairo_line_to(context, x, y);

  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(move_to, 0){
  cairo_t* context;
  double x, y;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_DOUBLE_ARG(args, x);
  DFSCH_DOUBLE_ARG(args, y);
  DFSCH_ARG_END(args);

  cairo_move_to(context, x, y);

  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(rectangle, 0){
  cairo_t* context;
  double x, y, width, height;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_DOUBLE_ARG(args, x);
  DFSCH_DOUBLE_ARG(args, y);
  DFSCH_DOUBLE_ARG(args, width);
  DFSCH_DOUBLE_ARG(args, height);
  DFSCH_ARG_END(args);

  cairo_rectangle(context, x, y, width, height);

  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(rel_curve_to, 0){
  cairo_t* context;
  double x1, y1, x2, y2, x3, y3;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_DOUBLE_ARG(args, x1);
  DFSCH_DOUBLE_ARG(args, y1);
  DFSCH_DOUBLE_ARG(args, x2);
  DFSCH_DOUBLE_ARG(args, y2);
  DFSCH_DOUBLE_ARG(args, x3);
  DFSCH_DOUBLE_ARG(args, y3);
  DFSCH_ARG_END(args);

  cairo_rel_curve_to(context, x1, y1, x2, y2, x3, y3);

  return NULL;
}

DFSCH_DEFINE_PRIMITIVE(rel_line_to, 0){
  cairo_t* context;
  double x, y;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_DOUBLE_ARG(args, x);
  DFSCH_DOUBLE_ARG(args, y);
  DFSCH_ARG_END(args);

  cairo_rel_line_to(context, x, y);

  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(rel_move_to, 0){
  cairo_t* context;
  double x, y;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_DOUBLE_ARG(args, x);
  DFSCH_DOUBLE_ARG(args, y);
  DFSCH_ARG_END(args);

  cairo_rel_move_to(context, x, y);

  return NULL;
}

/*
 * Transformations
 */
DFSCH_DEFINE_PRIMITIVE(translate, 0){
  cairo_t* context;
  double tx, ty;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_DOUBLE_ARG(args, tx);
  DFSCH_DOUBLE_ARG(args, ty);
  DFSCH_ARG_END(args);

  cairo_translate(context, tx, ty);

  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(scale, 0){
  cairo_t* context;
  double sx, sy;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_DOUBLE_ARG(args, sx);
  DFSCH_DOUBLE_ARG(args, sy);
  DFSCH_ARG_END(args);

  cairo_scale(context, sx, sy);

  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(rotate, 0){
  cairo_t* context;
  double angle;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_DOUBLE_ARG(args, angle);
  DFSCH_ARG_END(args);

  cairo_rotate(context, angle);

  return NULL;
}
DFSCH_DEFINE_PRIMITIVE(identity, 0){
  cairo_t* context;

  CAIRO_CONTEXT_ARG(args, context);
  DFSCH_ARG_END(args);

  cairo_identity_matrix(context);

  return NULL;
}



/*
 * Surfaces
 */

/*
 * Image surface
 */

DFSCH_DEFINE_PRIMITIVE(make_image_surface, 0){
  cairo_format_t format;
  int width;
  int height;
  
  CAIRO_FORMAT_ARG(args, format);
  DFSCH_LONG_ARG(args, width);
  DFSCH_LONG_ARG(args, height);
  DFSCH_ARG_END(args);

  return dfsch_ext_cairo_make_surface(cairo_image_surface_create(format, 
                                                                 width, 
                                                                 height),
                                      DFSCH_EXT_CAIRO_IMAGE_SURFACE_TYPE);
}

// . . .

#ifdef CAIRO_HAS_PNG_FUNCTIONS
/*
 * PNG support
 */

DFSCH_DEFINE_PRIMITIVE(write_surface_to_png, 0){
  cairo_surface_t* surface;
  char* filename;

  CAIRO_SURFACE_ARG(args, surface);
  DFSCH_STRING_ARG(args, filename);
  DFSCH_ARG_END(args);

  cairo_surface_write_to_png(surface, filename);

  return NULL;
}


#endif





void dfsch_module_cairo_register(dfsch_object_t* env){
  dfsch_provide(env, "cairo");

  dfsch_define_cstr(env, "cairo:<context>", DFSCH_EXT_CAIRO_CONTEXT_TYPE);
  dfsch_define_cstr(env, "cairo:<surface>", DFSCH_EXT_CAIRO_SURFACE_TYPE);
  dfsch_define_cstr(env, "cairo:<image-surface>", 
                    DFSCH_EXT_CAIRO_IMAGE_SURFACE_TYPE);

#ifdef CAIRO_HAS_PDF_SURFACE
  dfsch_define_cstr(env, "cairo:<pdf-surface>", 
                    DFSCH_EXT_CAIRO_PDF_SURFACE_TYPE);
#endif
#ifdef CAIRO_HAS_PS_SURFACE
  dfsch_define_cstr(env, "cairo:<ps-surface>", 
                    DFSCH_EXT_CAIRO_PS_SURFACE_TYPE);
#endif
#ifdef CAIRO_HAS_SVG_SURFACE
  dfsch_define_cstr(env, "cairo:<svg-surface>", 
                    DFSCH_EXT_CAIRO_SVG_SURFACE_TYPE);
#endif
#ifdef CAIRO_HAS_XLIB_SURFACE
  dfsch_define_cstr(env, "cairo:<xlib-surface>", 
                    DFSCH_EXT_CAIRO_XLIB_SURFACE_TYPE);
#endif

  dfsch_define_cstr(env, "cairo:make-context", 
                    DFSCH_PRIMITIVE_REF(make_context));
  dfsch_define_cstr(env, "cairo:destroy-context!", 
                    DFSCH_PRIMITIVE_REF(destroy_context));
  dfsch_define_cstr(env, "cairo:save-context!", 
                    DFSCH_PRIMITIVE_REF(save_context));
  dfsch_define_cstr(env, "cairo:restore-context!", 
                    DFSCH_PRIMITIVE_REF(restore_context));

  dfsch_define_cstr(env, "cairo:push-group!", 
                    DFSCH_PRIMITIVE_REF(push_group));
  dfsch_define_cstr(env, "cairo:pop-group!", 
                    DFSCH_PRIMITIVE_REF(pop_group));
  dfsch_define_cstr(env, "cairo:pop-group-to-source!", 
                    DFSCH_PRIMITIVE_REF(pop_group_to_source));

  dfsch_define_cstr(env, "cairo:set-source!", 
                    DFSCH_PRIMITIVE_REF(set_source));
  dfsch_define_cstr(env, "cairo:set-source-color!", 
                    DFSCH_PRIMITIVE_REF(set_source_color));
  dfsch_define_cstr(env, "cairo:set-source-surface!", 
                    DFSCH_PRIMITIVE_REF(set_source_surface));

  dfsch_define_cstr(env, "cairo:set-antialias!", 
                    DFSCH_PRIMITIVE_REF(set_antialias));
  dfsch_define_cstr(env, "cairo:set-line-width!", 
                    DFSCH_PRIMITIVE_REF(set_line_width));

  dfsch_define_cstr(env, "cairo:paint!", 
                    DFSCH_PRIMITIVE_REF(paint));
  dfsch_define_cstr(env, "cairo:fill!", 
                    DFSCH_PRIMITIVE_REF(fill));
  dfsch_define_cstr(env, "cairo:fill-preserve!", 
                    DFSCH_PRIMITIVE_REF(fill_preserve));
  dfsch_define_cstr(env, "cairo:stroke!", 
                    DFSCH_PRIMITIVE_REF(stroke));
  dfsch_define_cstr(env, "cairo:stroke-preserve!", 
                    DFSCH_PRIMITIVE_REF(stroke_preserve));

  dfsch_define_cstr(env, "cairo:current-point", 
                    DFSCH_PRIMITIVE_REF(current_point));
  dfsch_define_cstr(env, "cairo:new-path!", 
                    DFSCH_PRIMITIVE_REF(new_path));
  dfsch_define_cstr(env, "cairo:new-sub-path!", 
                    DFSCH_PRIMITIVE_REF(new_sub_path));
  dfsch_define_cstr(env, "cairo:close-path!", 
                    DFSCH_PRIMITIVE_REF(close_path));

  dfsch_define_cstr(env, "cairo:arc!", 
                    DFSCH_PRIMITIVE_REF(arc));
  dfsch_define_cstr(env, "cairo:arc-negative!", 
                    DFSCH_PRIMITIVE_REF(arc_negative));
  dfsch_define_cstr(env, "cairo:curve-to!", 
                    DFSCH_PRIMITIVE_REF(curve_to));
  dfsch_define_cstr(env, "cairo:line-to!", 
                    DFSCH_PRIMITIVE_REF(line_to));
  dfsch_define_cstr(env, "cairo:move-to!", 
                    DFSCH_PRIMITIVE_REF(move_to));
  dfsch_define_cstr(env, "cairo:rel-curve-to!", 
                    DFSCH_PRIMITIVE_REF(rel_curve_to));
  dfsch_define_cstr(env, "cairo:rel-line-to!", 
                    DFSCH_PRIMITIVE_REF(rel_line_to));
  dfsch_define_cstr(env, "cairo:rel-move-to!", 
                    DFSCH_PRIMITIVE_REF(rel_move_to));
  dfsch_define_cstr(env, "cairo:rectangle!", 
                    DFSCH_PRIMITIVE_REF(rectangle));



  dfsch_define_cstr(env, "cairo:make-image-surface", 
                    DFSCH_PRIMITIVE_REF(make_image_surface));

#ifdef CAIRO_HAS_PNG_FUNCTIONS
  dfsch_define_cstr(env, "cairo:write-surface-to-png", 
                    DFSCH_PRIMITIVE_REF(write_surface_to_png));
#endif


}
